#include "Precompiled.h"
#include "ActionTable.h"
#include "Track.h"
#include "Action.h"
#include "CommonFunc.h"


ActionTable::ActionTable() : m_nSortCol(-1), m_bSortColDec(false), m_nNumRows(0),
	m_mode(modeTrack)
{
}

ActionTable::~ActionTable()
{
	RemoveActions();
}

// called before doing search
void ActionTable::RemoveActions()
{
	map<int, Action*>::iterator iterMap; 
	for (iterMap = m_mapAction.begin(); iterMap != m_mapAction.end(); iterMap++) 
		delete iterMap->second;
	m_mapAction.clear();
	m_nNumRows = 0;
	m_nSortCol = -1;
	m_bSortColDec = false;
}

int ActionTable::GetNumberCols()
{
	return c_nLastCol+1;
}

const char* c_szTrackColHeader[c_nLastCol+1] = {"Action", "", "Title", "Artist", "Album", 
"Genre", "Length", "File size", "Track", "Year", "Path"};

const char* c_szPlaylistColHeader[c_nLastCol+1] = {"Action", "", "Playlist", "Tracks", "Status", 
"", "", "", "", "", "Path"};

wxString ActionTable::GetColLabelValue(int nCol)
{
	if (m_mode == modeTrack) return c_szTrackColHeader[nCol];
	return c_szPlaylistColHeader[nCol];
}
wxString ActionTable::GetRowLabelValue(int nRow)
{
	return wxString::Format("%d", nRow+1); 
}


void ActionTable::AddAction(Action* pAction)
{
	map<int, Action*>::iterator iterMap; 
	// every 2nd pair is a different colour
	if (pAction->IsPaired() && m_nNumRows > 0)
	{
		iterMap = m_mapAction.find(m_nNumRows-1); 
	// no row found -> must be the 2nd row of a label action
		if (iterMap == m_mapAction.end()) 
		{
			iterMap = m_mapAction.find(m_nNumRows-2); 
		}
		if (iterMap != m_mapAction.end() &&
			iterMap->second != pAction->GetDependAction())
		{
			pAction->SetEvenPair(!iterMap->second->GetEvenPair());
		}
	}
	m_mapAction[m_nNumRows] = pAction;
	m_nNumRows++;
	if (pAction->IsLabelAction()) m_nNumRows++;
}	// AddAction


// The grid calls GetValue and GetAttr on demand (i.e. only to display visible rows)
wxString ActionTable::GetValue(int nRow, int nCol)
{
	if (m_mapAction.size() == 0) return "";
	map<int, Action*>::iterator iterMap; 
	iterMap = m_mapAction.find(nRow); 
	int nTrackIX = 0;	// trackIX = 1 when it is the 2nd row in a label action
	// no row found -> must be the 2nd row of a label action
	if (iterMap == m_mapAction.end()) 
	{
		iterMap = m_mapAction.find(nRow-1); 
		wxASSERT(iterMap != m_mapAction.end());
		if (iterMap == m_mapAction.end()) return "";
		wxASSERT(iterMap->second->IsLabelAction());
		nTrackIX = 1;
	}
	return iterMap->second->GetValue(nCol, nTrackIX);
}


wxGridCellAttr* ActionTable::GetAttr(int nRow, int nCol, 
	wxGridCellAttr::wxAttrKind kind)
{
	if (m_mapAction.size() == 0) return NULL;
	map<int, Action*>::iterator iterMap; 
	iterMap = m_mapAction.find(nRow); 
	int nTrackIX = 0;
	// no row found -> must be the 2nd row of a lable action
	if (iterMap == m_mapAction.end()) 
	{
		iterMap = m_mapAction.find(nRow-1); 
		wxASSERT(iterMap != m_mapAction.end());
		if (iterMap == m_mapAction.end()) return NULL;
		wxASSERT(iterMap->second->IsLabelAction());
		nTrackIX = 1;
	}
	return iterMap->second->GetAttr(nCol, nTrackIX, m_attrCache);
}	// GetAttr


// SetValue called when the user changes an action
void ActionTable::SetValue(int nRow, int nCol, const wxString& strValue)
{
	wxASSERT(nCol == c_nActionCol);
	// first check that changed cell is within selection
	bool bRefresh = false;
	if (m_pGrid->IsInSelection(nRow, nCol))
	{
		unsigned int n;
		int nRowCnt;
		// First check rows selected by row header
		// (because stupid grid treats rows selected by row header differently)
		wxArrayInt listSelRow = m_pGrid->GetSelectedRows();
		for (n = 0; n < listSelRow.GetCount(); n++)
		{
			SetSingleValue(listSelRow.Item(n), nCol, strValue);
		}
		// Then check rows selected by dragging across cells
		wxGridCellCoordsArray blockTopLeft = m_pGrid->GetSelectionBlockTopLeft();
		wxGridCellCoordsArray blockBottomRight =  
			m_pGrid->GetSelectionBlockBottomRight();
		for (n = 0; n < blockTopLeft.GetCount(); n++ )
		{
			for (nRowCnt = blockTopLeft[n].GetRow(); 
				nRowCnt <= blockBottomRight[n].GetRow(); nRowCnt++)
			{
				SetSingleValue(nRowCnt, nCol, strValue);
			}
		}
		bRefresh = true;
	}
	else bRefresh = SetSingleValue(nRow, nCol, strValue);
	if (bRefresh) m_pGrid->ForceRefresh();
	if (m_nSortCol == nCol) 
	{
		m_nSortCol = -1;
		m_bSortColDec = false;
	}
}	// SetValue


// SetValue called when the user changes an action
bool ActionTable::SetSingleValue(int nRow, int nCol, const wxString& strValue)
{
	map<int, Action*>::iterator iterMap, iterMap2; 
	iterMap = m_mapAction.find(nRow); 
	if (iterMap == m_mapAction.end()) return false;	// user may have selected the 2nd row in a label change
	if (iterMap->second->IsDone()) return false;
	bool bRefresh = iterMap->second->SetAction(strValue);
	// label action pairs may be split, or merged 
	Action* pNewAction = iterMap->second->CheckSplit();
	if (pNewAction)
	{
		m_mapAction[nRow+1] = pNewAction;
		bRefresh = true;
	}
	else
	{
		iterMap2 = m_mapAction.find(nRow+1); 
		if (iterMap2 != m_mapAction.end() && 
			iterMap->second->MergeActions(iterMap2->second))
		{
			m_mapAction.erase(iterMap2);
			bRefresh = true;
		}
	}
	return bRefresh;
}	// SetValue


// GetActionsInOrder fills listAction with all actions in the order that
// they need to be exectuted
void ActionTable::GetActionsInOrder(ActionList& listAction) const
{
	Action* pAction;
	map<int, Action*>::const_iterator iterMap; 
	listAction.clear();
	for (int n = 0; n < m_nNumRows; n++)
	{
		iterMap = m_mapAction.find(n); 
		if (iterMap == m_mapAction.end()) continue;
		pAction = iterMap->second;
		listAction.push_back(pAction);
		if (pAction->GetDependAction())
		{
			if (pAction->IsSecond())
			{
				listAction[listAction.size()-1] = pAction->GetDependAction();
				listAction.push_back(pAction);
			}
			else listAction.push_back(pAction->GetDependAction());
			n++;
		}
	}
}	// GetActionsInOrder

int ActionTable::ActionsToDo() const
{
	int nActionsToDo = 0;
	map<int, Action*>::const_iterator iterMap; 
	for (iterMap = m_mapAction.begin(); iterMap != m_mapAction.end(); iterMap++) 
		if (iterMap->second->IsPending())
		{
			nActionsToDo++;
		}
	return nActionsToDo;
}	// ActionsToDo

void ActionTable::SetDefaultActions()
{
	if (m_pGrid->IsCellEditControlShown())	
	{
		m_pGrid->DisableCellEditControl();		
	}
	map<int, Action*>::iterator iterMap; 
	for (iterMap = m_mapAction.begin(); iterMap != m_mapAction.end(); iterMap++) 
	{
		iterMap->second->SetDefault();
	}
	m_pGrid->ForceRefresh();
}

void ActionTable::Sort(int nCol)
{
	int nActions = m_mapAction.size();
	if (nActions < 2 || nCol < c_nActionCol || nCol > c_nLastCol) return;
	// remember which action we have selected
	int nSelRow = -1;
	Action* pSelAction = NULL;
//	nSelRow = m_pGrid->GetGridCursorRow();
	wxArrayInt listSelRow = m_pGrid->GetSelectedRows();
	if (listSelRow.GetCount() > 0)
	{
		nSelRow = listSelRow.Item(0);
	}
	else
	{
		wxGridCellCoordsArray blockTopLeft = m_pGrid->GetSelectionBlockTopLeft();
		if (blockTopLeft.GetCount() > 0)
			nSelRow = blockTopLeft[0].GetRow(); 
	}
	if (nSelRow > -1)
	{
		map<int, Action*>::iterator iterMap, iterMap2; 
		iterMap = m_mapAction.find(nSelRow); 
		// check if it's the 2nd row in a pair
		if (iterMap == m_mapAction.end()) iterMap = m_mapAction.find(nSelRow-1); 
		if (iterMap != m_mapAction.end()) 
			pSelAction = iterMap->second;
	}
	
	// we must disable the edit control before we move any rows
	if (m_pGrid->IsCellEditControlShown())	
	{
		m_pGrid->EnableCellEditControl(false);
	}
	// first copy the map into a simple array of pointers
	Action** pListAction = new Action*[m_nNumRows*2];
	memset(pListAction, 0, sizeof(Action*)*m_nNumRows*2);
	int n = 0;
	map<int, Action*>::iterator iterMap; 
	int nRow = m_nNumRows-1;
	while (nRow >= 0)
	{
		iterMap = m_mapAction.find(nRow); // copy reversed order
		if (iterMap == m_mapAction.end()) 
		{
			nRow--;
			iterMap = m_mapAction.find(nRow); 
			wxASSERT(iterMap != m_mapAction.end());
		}
		pListAction[n] = iterMap->second;
		nRow--;
		if (iterMap->second->GetDependAction() != NULL)
		{
			// as we're copying backwards we will have put the 2nd
			// pair into the first location
			pListAction[n+1] = pListAction[n];
			pListAction[n] = iterMap->second->GetDependAction();
			nRow--;
		}
		n += 2; 
	}
	int nListSize = n/2;
	// if the column is already sorted then we do not need to do anything
	// becuase we copied in reverse order they will be automatically reversed
	if (m_nSortCol == nCol) m_bSortColDec = !m_bSortColDec;
	else 
	{
		m_bSortColDec = false;
		m_nSortCol = nCol;
		ShellSort(pListAction, nListSize, nCol);
	}
	nRow = 0;
	m_mapAction.clear();
	// copy back into the map
	for (n = 0; n < nListSize*2; n+=2) 
	{
		m_mapAction[nRow] = pListAction[n];
		nRow++;
		if (pListAction[n]->IsPaired())
		{
			if (pListAction[n]->GetDependAction() != NULL)
			{
				wxASSERT(pListAction[n+1] != NULL);
				m_mapAction[nRow] = pListAction[n+1];
			}
			nRow++;
			if (n == 0) pListAction[n]->SetEvenPair(true);
			else 
			{
				pListAction[n]->SetEvenPair(!pListAction[n-2]->GetEvenPair());
			}
		}
	}
	wxASSERT(nRow == m_nNumRows);
	delete [] pListAction;
	ScrollToAction(pSelAction);
	m_pGrid->ForceRefresh();
	// it would be nice to select the previous selected row, but we can't
	// because we're in the middle of a click event
//	if (nNewSelRow > 0) m_pGrid->SelectRow(nNewSelRow);
}	// Sort

void ActionTable::ShellSort(Action** pListAction, int nListSize, int nCol)
{
	Action *pTemp, *pTemp2;
	bool bSwapped = true;
	int nCompareDist = nListSize;
	int n;
	while (bSwapped || nCompareDist>1)      
	{
		bSwapped = false;           
		nCompareDist = (nCompareDist+1) / 2;
		for (n = 0; n < nListSize - nCompareDist; n++)
		{
			if (pListAction[(n+nCompareDist)*2]->ColIsLessThan(nCol, pListAction[n*2]))
			{
				pTemp = pListAction[(n + nCompareDist)*2];      
				pTemp2 = pListAction[(n + nCompareDist)*2+1];      
				pListAction[(n + nCompareDist)*2] = pListAction[n*2];
				pListAction[(n + nCompareDist)*2+1] = pListAction[n*2+1];
				pListAction[n*2] = pTemp;
				pListAction[n*2+1] = pTemp2;
				bSwapped = true;                  
			}
		}
	}
}

// scrolls the row where pAction is so it's visble
int  ActionTable::ScrollToAction(const Action* pAction)
{
	if (pAction == NULL) return -1;
	map<int, Action*>::const_iterator iterMap; 
	for (iterMap = m_mapAction.begin(); iterMap != m_mapAction.end(); iterMap++) 
	{
		if (pAction == iterMap->second)
		{
			m_pGrid->MakeCellVisible(iterMap->first, c_nActionCol);
			return iterMap->first;
		}
	}
	return -1;
}
