#include "Precompiled.h"
#include "Playlist.h"
#include "PlaylistAction.h"
#include "BaseNomad.h"
#include "Config.h"
#include "CellAttrCache.h"
#include "GridCellChoiceRenderer.h"
#include "FastComboEditor.h"
#include "Track.h"


PlaylistAction::PlaylistAction(const Config& config) : Action(config),
	m_pDependAction(NULL)	
{
	m_pPlaylist = NULL;
}

PlaylistAction::~PlaylistAction()
{
}

bool PlaylistAction::DoAction(BaseNomad* pNomad)
{
	wxLogNull logNo;				
	bool bResult = false;
	m_bWaitForCompletion = false;
	m_bExecuting = true;
	switch(m_action)
	{
		case PC2Nomad:
			// make a copy of m_pPlaylist, because this object will be destroyed by
			// ActionTable
			if (!m_pPlaylist->CanTransfer())
			{
				m_strError = "Cannot transfer playlist with no files!";
				break;
			}
			bResult = pNomad->CreatePlaylist(new Playlist(*m_pPlaylist));
			if (!bResult)
				m_strError = pNomad->GetLastErrorStr();
			break;
		case Nomad2PC:
			bResult = m_pPlaylist->WriteM3uFile(m_config.GetUploadDir());
			if (!bResult)
				m_strError = wxSysErrorMsg(wxSysErrorCode());
			break;
		case Delete:
			if (m_pPlaylist->IsNomad())
			{
				bResult = pNomad->DeletePlaylist(m_pPlaylist);
				if (!bResult)
					m_strError = pNomad->GetLastErrorStr();
			}
			else
			{
				bResult = MoveToReCycleBin(m_pPlaylist->GetFileSpec());
				if (!bResult)
					m_strError = wxSysErrorMsg(wxSysErrorCode());
			}
			break;
		default:
			break;
	}
	if (bResult) m_bDone = true;
	if (HadError()) wxLogMessage(m_strError);
	return bResult;
}	// DoAction


// returns true if the other row in a pair item needs to be refreshed
bool PlaylistAction::SetAction(ActionEnum action)
{
	m_action = action;
	m_bUserModified = true;
	if (m_pAttr) m_pAttr->SetTextColour(*wxBLUE);
	// Prevent uploading/downloading of a duplicate track by making the
	// twin track action delete
	if (m_pDependAction)	// if there is a dependant action, then it must have the same name
	{
		if (m_action == PC2Nomad || m_action == Nomad2PC)
		{
			m_pDependAction->m_action = Delete;
			m_pDependAction->m_bUserModified = true;
			return true;
		}
		if (m_action == noAction && m_pDependAction->m_action == PC2Nomad)
		{
			m_pDependAction->m_action = noAction;
			m_pDependAction->m_bUserModified = true;
			return true;
		}
	}
	return false;
}



// return the estimated time needed to do the action
// in terms of bytes to transfer
int PlaylistAction::GetTimeNeeded() const
{
	return 5000;
}

wxString PlaylistAction::GenErrorMsg() const
{
	wxString strErr = "Operation: " + c_strAction[m_action];
	strErr += "\n";
	strErr += "Playlist: " + m_pPlaylist->GetName() + "\n";
	strErr += "Error message: ";
	strErr += m_strError;
	return strErr;
}

wxString PlaylistAction::GetDisplayStr() const
{
	wxString str = c_strAction[m_action];
	str += "Playlist: ";
	str += m_pPlaylist->GetName();
	return str;
}



void PlaylistAction::SetDefault() 
{
	if (m_bDone || HadError() || m_bUserModified) return;
	if (!m_pPlaylist->CanTransfer())
	{
		SetChoice(0, Delete);
		SetChoice(1, noAction);
		m_nChoices = 2;
		m_action = noAction;
		return;
	}
	SetChoice(0, Delete);
	SetChoice(2, noAction);
	m_nChoices = 3;
	if (m_pPlaylist->IsNomad())
	{
		SetChoice(1, Nomad2PC);
		m_action = m_config.GetActionDefault(c_nDefMissingOnPC);
	}
	else
	{
		SetChoice(1, PC2Nomad);
		m_action = m_config.GetActionDefault(c_nDefMissingOnNomad);
	}
	if (m_pDependAction && !m_pDependAction->m_bUserModified)
	{
		if (m_action == PC2Nomad || m_action == Nomad2PC)
		{
			m_pDependAction->m_action = Delete;
			return;
		}
		ActionEnum depAction;
		if (m_pDependAction->m_pPlaylist->IsNomad())
			depAction = m_config.GetActionDefault(c_nDefMissingOnPC);
		else
			depAction = m_config.GetActionDefault(c_nDefMissingOnNomad);
		if (!m_pDependAction->m_pPlaylist->CanTransfer()) depAction = noAction;
		if (m_action == noAction && (depAction == PC2Nomad || depAction == Nomad2PC))
		{
			m_action = Delete;
		}
	}
} // Set Default


// trackIX = 1 when it is the 2nd row in a label action
wxString PlaylistAction::GetValue(int nCol, int nTrackIX) const
{
	if (nCol == c_nActionCol) 
	{
		if (m_bExecuting) return "";
		if (HadError()) 
		{
			wxString strErr = "Error: " + m_strError;
			return strErr;
		}
		if (!m_bDone) return c_strAction[m_action];
		return c_strActionDone[m_action];
	}
	return m_pPlaylist->GetValue(nCol);
}


wxGridCellAttr* PlaylistAction::GetAttr(int nCol, int nTrackIX, CellAttrCache& attrCache) const
{
	if (nCol == c_nIconCol) 
	{
		return attrCache.Get((CellTypeUsed)(ctIcon+GetColourIndex() ) );
	}
	if (nCol == c_nActionCol)
	{
		if (m_bExecuting) return attrCache.Get((CellTypeUsed)ctNormal);
		if (m_bDone || HadError()) 
			return attrCache.Get((CellTypeUsed)(ctNormal+GetColourIndex()));
		if (m_pAttr) 
		{
			m_pAttr->IncRef();
			return m_pAttr;
		}
		m_pAttr = new wxGridCellAttr;
		m_pAttr->IncRef();
		m_pAttr->SetBackgroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_WINDOW));
		wxLayoutAlignment border = wxLAYOUT_NONE;
		m_pAttr->SetRenderer(new wxGridCellChoiceRenderer(border));
		m_pAttr->SetEditor(new wxFastComboEditor(m_nChoices, m_strChoices, TRUE));
		return m_pAttr;
	}
	return attrCache.Get((CellTypeUsed)(ctNormal+GetColourIndex()));
}	// GetAttr


// for icon column we should sort according to pc, nomad & pair
bool PlaylistAction::ColIsLessThan(int nCol, const Action* pAction) const
{
	PlaylistAction* pPLAction = (PlaylistAction*)pAction;
	if (nCol == c_nActionCol) 
	{
		if (c_strAction[m_action] == c_strAction[pPLAction->GetAction()]) 
			return m_pPlaylist->ColIsLessThan(c_nTitleCol, pPLAction->m_pPlaylist);
		return c_strAction[m_action] < c_strAction[pPLAction->GetAction()]; 
	}
	if (nCol == c_nIconCol) 
	{
		if (GetTypeOrder() == pPLAction->GetTypeOrder())
			return m_pPlaylist->ColIsLessThan(c_nTitleCol, pPLAction->m_pPlaylist);
		return GetTypeOrder() < pPLAction->GetTypeOrder();
	}
	return m_pPlaylist->ColIsLessThan(nCol, pPLAction->m_pPlaylist);
}

int PlaylistAction::GetTypeOrder() const	
{
	if (m_pDependAction != NULL) return 3;
	if (m_pPlaylist->IsNomad()) return 2;
	return 1;
}

bool PlaylistAction::IsPaired() const	
{
	return m_pDependAction != NULL;
}


