#include "Precompiled.h"
#include "Track.h"
#include "EzTag.h"
#include "EzGrid.h"
#include <math.h>
#include <wx/filename.h>
#include "BaseNomad.h"
#include "CommonFunc.h"
#include "Config.h"

#ifdef __WINDOWS__
#include "libmpg123\mpg123.h"
#else
#include "libmpg123/mpg123.h"
#endif


using namespace std;

// These are used because you cannot set a empty string for any attribute
// on the nomad
const wxString c_strBlankAttrib = "<UNKNOWN>";
const wxString c_strBlankGenre = "(None)";


// array to translate a column location to the Attribute 
const int c_nColToAttrib[c_nLastCol+1] = {-1, -1, taTitle, taArtist, taAlbum, 
	taGenre, taLength, taFileSize, taTrackNum, taYear, taFileSpec};

const int c_nCompareExtra[c_nMaxCompareExtra] = {taGenre, taTrackNum, taYear};

int Track::m_nCountMissingTags = 0;

Track::Track() : m_nFileSize(-1), m_nTrackLength(-1), m_nTrackNum(0), m_nYear(0), 
	m_nItemID(0), m_bFaulty(false), m_timeFileModified(0), m_bMissingTag(false) 
{
}

Track::~Track()
{
}


// Loads the tag data from the file
bool Track::Load(const wxString& strFileSpec)
{
	m_strAttrib[taFileSpec] = strFileSpec;
	EzTag id3tag;
	int nSize = id3tag.Link(strFileSpec, ID3TT_ID3);
//	if (id3tag.HasV2Tag() && id3tag.HasV1Tag() ) 
//		id3tag.Link(strFileSpec, ID3TT_ID3V1);
	m_timeFileModified = wxFileModificationTime(strFileSpec);
	SetAttrib(taTitle, id3tag.GetField(ID3FID_TITLE));
	SetAttrib(taArtist, id3tag.GetField(ID3FID_LEADARTIST));
	SetAttrib(taAlbum, id3tag.GetField(ID3FID_ALBUM));
	SetAttrib(taGenre, id3tag.GetField(ID3FID_CONTENTTYPE));
	// id3v1 tags use a number for the genre, we must translate this number
	if (m_strAttrib[taGenre].Length() > 2 && 
		m_strAttrib[taGenre].GetChar(0) == '(') 
	{
		int nPos = m_strAttrib[taGenre].Find(')');
		if (nPos != -1)
		{
			wxString str = m_strAttrib[taGenre].Mid(1, nPos-1);
			long nGenreIX = 0;
			if (str.ToLong(&nGenreIX) && nGenreIX >= 0 && nGenreIX < ID3_NR_OF_V1_GENRES) 
			{
				m_strAttrib[taGenre] = ID3_v1_genre_description[nGenreIX];
			}
		}
	}
	m_strAttrib[taCodec] = "mp3";
	m_nTrackNum = atoi(id3tag.GetField(ID3FID_TRACKNUM));
	m_nYear = atoi(id3tag.GetField(ID3FID_YEAR));
	m_nTrackLength = -1;		
	FILE *file;
	if ((file = fopen(strFileSpec, "rb")) != NULL)
	{
		int nLength = mpg123_get_song_time(file);
		if (nLength == 0)
		{
			wxLogMessage("file %s, can't get track length", 
				(const char*)strFileSpec);
		}
		else m_nTrackLength = (int)((double)nLength/1000.0 + 0.5);
		fclose(file);
	}
	m_nFileSize = id3tag.GetFileSize();
	if (m_nTrackLength == -1 && (m_nFileSize <= 30000 ||
		m_strAttrib[taTitle].IsEmpty())) 
	{
		wxLogMessage("Can't get mp3 header for");
		wxLogMessage(strFileSpec);
		SetAttrib(taArtist, "Faulty track!!");
		m_bFaulty = true;
	}
	if (m_strAttrib[taTitle].IsEmpty() && m_strAttrib[taArtist].IsEmpty() 
		&& m_strAttrib[taAlbum].IsEmpty()) 
	{
		m_nCountMissingTags++;
		SetAttrib(taTitle, wxString::Format("Missing tag %d", m_nCountMissingTags));
		m_bMissingTag = true;
	}
	return true;
}	//	Load

// update mechanism of grid calls this function indirectly 
wxString Track::GetValue(int nCol) const
{
	if (nCol == c_nIconCol) 
	{
		if (IsNomad())
			return "IDI_NOMAD";
		else
			return "IDI_PC";
	}
	wxASSERT(nCol >= c_nTitleCol && nCol <= c_nFileSpecCol && c_nColToAttrib[nCol] != -1);
	return GetAttribAsStr(c_nColToAttrib[nCol]);
}

// Returns true if the attribute specified by nCol is less than the one in pTrack
bool Track::ColIsLessThan(int nCol, const Track* pTrack) const
{
	wxASSERT(nCol >= c_nTitleCol && nCol <= c_nFileSpecCol && c_nColToAttrib[nCol] != -1);
	int n = c_nColToAttrib[nCol];
	if (n >= taFileSpec && n < taMaxStrAttribs) 
		return m_strAttrib[n] < pTrack->GetAttrib((trackAttribStr)n);
	if (n == taFileSize) return m_nFileSize < pTrack->GetFileSize();
	if (n == taLength) return m_nTrackLength < pTrack->GetTrackLength();
	return false;
}

// used to uniquely identify a track on the nomad
wxString Track::GetKey(const Config* pConfig) const
{
	wxString strKey;
	strKey = m_strAttrib[taTitle]+m_strAttrib[taArtist]+m_strAttrib[taAlbum];
	if (pConfig)
	{
		for (int n = 0; n < 3; n++)
		{
			if (pConfig->GetCompareExtra(n))
				strKey += GetAttribAsStr(c_nCompareExtra[n]);
		}
	}
	return strKey;
}

// used for label matching
wxString Track::GetShortKey(int nAttribLen) const
{
	wxString strKey;
	strKey = m_strAttrib[taTitle].Left(nAttribLen) +
		m_strAttrib[taArtist].Left(nAttribLen) + 
		m_strAttrib[taAlbum].Left(nAttribLen);
	return strKey;
}

// this allows for a track being different size because the label was updated
const int c_nFileSizeTol = 0x0ffff;	// 64kB

// account for size of id3tag 
bool Track::SameSize(const Track* pTrack) const
{
	return fabs((double)(m_nFileSize - pTrack->GetFileSize())) < c_nFileSizeTol;
}	// SameSize

// account for size of id3tag 
bool Track::SameSizeAndLen(const Track* pTrack) const
{
	bool bSimiliarSize = SameSize(pTrack);
	int nLenDiff = (int)fabs((double)(m_nTrackLength - pTrack->GetTrackLength()));
	// As we are only updating the label we can ignore cases where the track len
	// was not found
	if (m_nTrackLength == -1 || pTrack->GetTrackLength() == -1) nLenDiff = 0;
	int nMinDiff = 1;
	if (m_nTrackLength > 800) nMinDiff = 2;
	bool bSameLen = nLenDiff <= nMinDiff;  
	return bSimiliarSize && bSameLen;
}	// SameSize

bool Track::SameKey(const Track* pTrack) const
{
	return GetKey() == pTrack->GetKey();
}

bool Track::EqualNoCase(const Track* pTrack) const
{
	return GetKey().CmpNoCase(pTrack->GetKey()) == 0;
}

// IsSimiliar returns true if the tracks are nearly the same
// nSimilarityLevel indicates the formula to use
// the higher the number the easier it is to find a match
bool Track::IsSimiliar(const Track* pTrack, int nSimilarityLevel) const
{
	if (m_bFaulty || pTrack->IsFaulty() || m_bMissingTag || pTrack->MissingTag()) 
		return false;
	bool bSimiliarSize = SameSize(pTrack);
	int nLenDiff = (int)fabs((double)(m_nTrackLength - pTrack->GetTrackLength()));
	// As we are only updating the label we can ignore cases where the track len
	// was not found
	if (m_nTrackLength == -1 || pTrack->GetTrackLength() == -1) nLenDiff = 0;
	bool bSameLen = nLenDiff < 2;  
	switch (nSimilarityLevel)
	{
		case 1:
			return bSimiliarSize && bSameLen &&
				GetShortKey(15).CmpNoCase(pTrack->GetShortKey(15)) == 0;
		case 2:
			return bSimiliarSize && bSameLen &&
				!m_strAttrib[taTitle].IsEmpty() &&
				m_strAttrib[taTitle].CmpNoCase(pTrack->GetAttrib(taTitle)) == 0;
		case 3:		
			return bSimiliarSize && bSameLen &&
				!m_strAttrib[taTitle].IsEmpty() &&
				m_strAttrib[taArtist].CmpNoCase(pTrack->GetAttrib(taArtist)) == 0 &&
				m_strAttrib[taAlbum].CmpNoCase(pTrack->GetAttrib(taAlbum)) == 0;
		case 4:
			return bSimiliarSize && bSameLen &&
				GetShortKey(6).CmpNoCase(pTrack->GetShortKey(6)) == 0;
		case 5:
			if (bSimiliarSize && bSameLen)
			{
				// Note: a weighted edit distance of 1.0 means all the characters are different!
				double dWeighted = WeighedEditDistance(GetKey(), pTrack->GetKey());
				if ( dWeighted > 2.0) return true;
				if ( dWeighted < 1.2) return false;
				if (WeighedEditDistance(GetShortKey(10), pTrack->GetShortKey(10)) 
					> 3.0) return true;
				if (WeighedEditDistance(m_strAttrib[taTitle]+m_strAttrib[taArtist], 
					pTrack->GetAttrib(taTitle) + pTrack->GetAttrib(taArtist)) > 4.5)
					return true;
			}
			return false;
		case 6:
		// special case when file sizes are exactly equal the matching is the least strict
			if (m_nFileSize == pTrack->GetFileSize() && bSameLen)
			{
				if (!m_strAttrib[taTitle].IsEmpty() &&	
					m_strAttrib[taTitle].Left(10).CmpNoCase(pTrack->GetAttrib(taTitle).Left(10)) == 0)
					return true;
				if (WeighedEditDistance(GetShortKey(10), pTrack->GetShortKey(10)) 
					> 1.5) return true;
			}
			if (fabs((double)(m_nFileSize - pTrack->GetFileSize())) < 2000.0 && bSameLen)
			{
				double dWeighted = WeighedEditDistance(GetKey(), pTrack->GetKey());
				if ( dWeighted > 1.9) return true;
				if ( dWeighted < 1.2) return false;
				if (WeighedEditDistance(GetShortKey(10), pTrack->GetShortKey(10)) 
					> 2.4) return true;
				if (WeighedEditDistance(m_strAttrib[taTitle]+m_strAttrib[taArtist], 
					pTrack->GetAttrib(taTitle) + pTrack->GetAttrib(taArtist)) 
					> 2.9) return true;
			}
			return false;
		case c_nSameLabelsNoCase:
			return EqualNoCase(pTrack);
		default:
			wxASSERT(false);
			return false;
	}
} // IsSimiliar

// returns full display name for transfer dialog
wxString Track::GetDisplayName() const
{
	if (m_strAttrib[taFileSpec].IsEmpty())
	{
		wxString str;
		str = m_strAttrib[taTitle] + " - ";
		str += m_strAttrib[taArtist] + " - ";
		str += m_strAttrib[taAlbum];
		return str;
	}
	return m_strAttrib[taFileSpec];
}

// Dumps itself to file
void Track::Dump(wxFile& fileOut) const
{
	if (m_bFaulty || m_bMissingTag) return;
	wxString str;
	for (int nAttrib = taFileSpec; nAttrib < taTotalAttribs; nAttrib++)
	{
		str += GetAttribAsStr(nAttrib) + '\t';
	}
	str += wxString::Format("%ld", m_timeFileModified) + "\n";
	fileOut.Write(str);
}

// parses Track data from memory buffer. Each attribute is
// delimited by a tab character
const char* Track::Parse(const char* pCurr, const char* pEnd) 
{
	int nAttrib = 0;
	const char* pStart;
	do
	{
		pStart = pCurr;
		while (*pCurr != '\t' && *pCurr != '\n' && pCurr != pEnd) 
		{
			pCurr++;
		}
		wxString str(pStart, pCurr-pStart);
		if (nAttrib >= taFileSpec && nAttrib < taMaxStrAttribs) 
		{
			m_strAttrib[nAttrib] = str;
		}
		else
		{
			int nNum = atoi(str);
			if (!str.IsNumber()) 
			{
				wxLogMessage("Error expection number in cache file");
				return NULL;
			}
			if (nAttrib == taFileSize) m_nFileSize = nNum;
			if (nAttrib == taLength) m_nTrackLength = nNum;
			if (nAttrib == taTrackNum) m_nTrackNum = nNum;
			if (nAttrib == taYear) m_nYear = nNum;
			if (nAttrib ==	taTotalAttribs) m_timeFileModified = (time_t)nNum;
		}
		nAttrib++;
		if (*pCurr == '\n' || pCurr == pEnd || nAttrib > taTotalAttribs+1) break;
		pCurr++;
	} while (true);
	if (nAttrib != taTotalAttribs+1)	
		return NULL;
	return pCurr+1;
}	// Parse


// When coping nomad files this generates the file name
wxFileName Track::GenerateFileSpec(const wxString& strPrimaryDir) const
{
	wxFileName fileSpec;
	fileSpec.AssignDir(strPrimaryDir);
	if (!m_strAttrib[taArtist].IsEmpty()) 
		fileSpec.AppendDir(StripDangerous(m_strAttrib[taArtist]));
	if (!m_strAttrib[taAlbum].IsEmpty()) 
	fileSpec.AppendDir(StripDangerous(m_strAttrib[taAlbum]));
	fileSpec.SetName(StripDangerous(m_strAttrib[taTitle]));
	fileSpec.SetExt(m_strAttrib[taCodec]);
	wxASSERT(fileSpec.IsOk());
	return fileSpec;
}


void Track::SetTag(const Track* pTrack)
{
	wxASSERT(!IsNomad());
	EzTag ezTag;
	ezTag.Link(m_strAttrib[taFileSpec]);	
	for (int nAttrib = taTitle; nAttrib <= taGenre; nAttrib++)
	{
		ezTag.UpdateTag(nAttrib, pTrack->GetAttrib((trackAttribStr)nAttrib));
	}
	wxString str;
	if (pTrack->GetTrackNum() > 0)
	{
		ezTag.UpdateTag(taTrackNum, pTrack->GetAttribAsStr(taTrackNum));
	}
	if (pTrack->GetYear() > 0)
	{
		ezTag.UpdateTag(taYear, pTrack->GetAttribAsStr(taYear));
	}
	ID3_TagType tagType = ID3TT_ID3V2;
	if (ezTag.HasV1Tag()) tagType = ID3TT_ID3;
	ezTag.Update((unsigned short)tagType);
}

void Track::SetTag(const wxString& strFileSpec)
{
	EzTag ezTag;
	ezTag.Link(strFileSpec, (flags_t)ID3TT_ALL);	
	for (int nAttrib = taTitle; nAttrib <= taGenre; nAttrib++)
	{
		ezTag.UpdateTag(nAttrib, m_strAttrib[nAttrib]);
	}
	if (m_nTrackNum > 0)
	{
		ezTag.UpdateTag(taTrackNum, GetAttribAsStr(taTrackNum));
	}
	if (m_nYear > 0)
	{
		ezTag.UpdateTag(taYear, GetAttribAsStr(taYear));
	}
	ID3_TagType tagType = ID3TT_ID3V2;
	if (ezTag.HasV1Tag()) tagType = ID3TT_ID3;
	ezTag.Update((unsigned short)tagType);
}

// returns the attribute without allowing any blanks
wxString Track::GetFilledAttrib(int n) const	 
{
	wxString strAttrib = m_strAttrib[n];
	if (strAttrib.IsEmpty()) 
	{
		if (n == taGenre)
			strAttrib = c_strBlankGenre;
		else if (n == taTitle)
		{
			wxASSERT(!m_strAttrib[taFileSpec].IsEmpty());
			wxFileName fn(m_strAttrib[taFileSpec]);
			strAttrib = fn.GetName();
		}
		else
			strAttrib = c_strBlankAttrib;
	}
	return strAttrib;
}

// returns any attribute as a string (including number attributes)
wxString Track::GetAttribAsStr(int n) const
{
	wxASSERT(n >= taFileSpec && n < taTotalAttribs);
	if (n >= taFileSpec && n < taMaxStrAttribs) return m_strAttrib[n];
	if (n == taFileSize) return wxString::Format("%d", m_nFileSize);
	if (n == taLength) return wxString::Format("%d", m_nTrackLength);
	if (n == taTrackNum) return wxString::Format("%d", m_nTrackNum);
	if (n == taYear) return wxString::Format("%d", m_nYear);
	wxString str;
	return str;
}

// Sets the attribute translating <Unknown> as a blank, etc
void Track::SetAttrib(trackAttribStr attrib, const wxString& strAttrib)
{
	// this is the maximum allowed size on the nomad
	m_strAttrib[attrib] = strAttrib.Left(99);		
	if (attrib == taCodec) m_strAttrib[attrib].MakeLower();
	if ((attrib == taGenre && m_strAttrib[attrib].CmpNoCase(c_strBlankGenre) == 0) ||
		(attrib != taGenre && m_strAttrib[attrib].CmpNoCase(c_strBlankAttrib) == 0) )
	{
		m_strAttrib[attrib].Empty();
	}
}


