#include "Precompiled.h"
#include "TrackAction.h"
#include "BaseNomad.h"
#include "Track.h"
#include "Config.h"
#include "CellAttrCache.h"
#include "GridCellChoiceRenderer.h"
#include "FastComboEditor.h"


TrackAction::TrackAction(const Config& config) : Action(config), m_pDependAction(NULL)
{
	m_pTrack[0] = NULL;
	m_pTrack[1] = NULL; 
}

TrackAction::~TrackAction()
{
}

bool TrackAction::DoAction(BaseNomad* pNomad)
{
	wxLogNull logNo;	// Make sure no error messages pop up!				
	bool bResult = false;
	m_bWaitForCompletion = false;
	m_bExecuting = true;
	switch(m_action)
	{
		case PC2Nomad:
			if (m_pTrack[0]->GetTrackLength() <= 0)
			{
				m_strError = "Missing track length info!";
				break;
			}
			bResult = pNomad->SendItem(m_pTrack[0]);
			if (bResult) m_bWaitForCompletion = true;
			else m_strError = pNomad->GetLastErrorStr();
			break;
		case Nomad2PC:
			m_fileDownLoaded = m_pTrack[0]->GenerateFileSpec(m_config.GetUploadDir());
			// create directory if it doesn't exist
			if (!m_fileDownLoaded.DirExists())
			{
				if (!m_fileDownLoaded.Mkdir(
					m_fileDownLoaded.GetPath(wxPATH_GET_VOLUME | wxPATH_GET_SEPARATOR), 
					0777, wxPATH_MKDIR_FULL))
				{
					m_strError = wxSysErrorMsg(wxSysErrorCode());
					break;
				}
			}
			else wxRemoveFile(m_fileDownLoaded.GetFullPath());	// just in case
			bResult = pNomad->GetItem(m_pTrack[0], m_fileDownLoaded);
			if (bResult) m_bWaitForCompletion = true;
			else m_strError = pNomad->GetLastErrorStr();
			break;
		case Delete:
			if (m_pTrack[0]->IsNomad())
			{
				bResult = pNomad->DeleteItem(m_pTrack[0]);
				if (!bResult) m_strError = pNomad->GetLastErrorStr();
			}
			else
			{
				bResult = MoveToReCycleBin(m_pTrack[0]->GetAttrib(taFileSpec));
				if (!bResult)
					m_strError = wxSysErrorMsg(wxSysErrorCode());
			}
			if (bResult) m_bDone = true;
			break;
		case Label2Nomad:
			bResult = pNomad->SendLabel(m_pTrack[0], m_pTrack[1]);
			if (bResult) m_bDone = true;
			else m_strError = pNomad->GetLastErrorStr();
			break;
		case Label2PC:
			m_pTrack[0]->SetTag(m_pTrack[1]);
			m_bDone = true;
			break;
		default:
			break;
	}
	if (HadError()) wxLogMessage(m_strError);
	return bResult;
}	// DoAction

bool TrackAction::DoPostAction(unsigned int nItemID)
{
	if (HadError()) 
	{
		m_bExecuting = false;
		return false;
	}
	if (m_action == Nomad2PC && m_fileDownLoaded.IsOk())
	{
		m_pTrack[0]->SetTag(m_fileDownLoaded.GetFullPath());
	}
	if (m_action == PC2Nomad && m_config.NeedPlaylists()) 
		m_pAddToPlaylist = m_pTrack[0];
	else m_pAddToPlaylist = NULL;
	if (m_action == PC2Nomad && nItemID != 0) 
		m_pTrack[0]->SetItemID(nItemID);;
	m_bDone = true;
	m_bWaitForCompletion = false;
	m_bExecuting = false;
	return true;
}

// returns true if the other row in a pair item needs to be refreshed
bool TrackAction::SetAction(ActionEnum action)
{
	m_action = action;
	m_bUserModified = true;
	if (m_pAttr) m_pAttr->SetTextColour(*wxBLUE);
	// Prevent uploading/downloading of a duplicate track by making the
	// twin track action delete
	if (m_pDependAction)
	{
		if ((m_action == PC2Nomad || m_action == Nomad2PC) &&
			m_pTrack[0]->EqualNoCase(m_pDependAction->m_pTrack[0]))
		{
			m_pDependAction->m_action = Delete;
			m_pDependAction->m_bUserModified = true;
			return true;
		}
		if ((m_action == noAction && m_pDependAction->m_action == PC2Nomad) &&
			m_pTrack[0]->EqualNoCase(m_pDependAction->m_pTrack[0]))
		{
			m_pDependAction->m_action = noAction;
			m_pDependAction->m_bUserModified = true;
			return true;
		}
	}
	return false;
}

// CheckSplit: checks if we create a new action item for the 2nd track
// of a relabel action
Action* TrackAction::CheckSplit()
{
	TrackAction* pAction2 = NULL;
	if (m_pTrack[1] != NULL && 
			m_action != Label2Nomad && m_action != Label2PC)
	{
		pAction2 = new TrackAction(m_config);
		pAction2->SetTrack(0, m_pTrack[1]);
		m_pTrack[1] = NULL;
		// force a dependancy so they stick together, even through a sort
		pAction2->SetDependAction(this);
		m_pDependAction = pAction2;
		pAction2->SetDefault();
	}
	return pAction2;
}

// deletes the dependant action 
bool TrackAction::MergeActions(Action* pNextAction) 
{
	if (m_pTrack[1] == NULL && 
			(m_action == Label2Nomad || m_action == Label2PC))
	{
		m_pTrack[1] = ((TrackAction*)pNextAction)->m_pTrack[0];
		delete pNextAction;
		m_pDependAction = NULL;
		return true;
	}
	return false;
}


// return the estimated time needed to do the action
// in terms of bytes to transfer
int TrackAction::GetTimeNeeded() const
{
	switch(m_action)
	{
		case Nomad2PC:
			// add more because need to re-label after copy
			return m_pTrack[0]->GetFileSize()+50000;
		case PC2Nomad:
			return m_pTrack[0]->GetFileSize();
		case Delete:
			if (m_pTrack[0]->IsNomad())
				return 100000;
			else
				return 10000;
		case Label2Nomad:
				return 5000;
		case Label2PC:
				return 50000;
		default:
			return 0;
	}
}

wxString TrackAction::GenErrorMsg() const
{
	wxString strErr = "Operation: " + c_strAction[m_action];
	strErr += "\n";
	strErr += "Track: " + m_pTrack[0]->GetDisplayName() + "\n";
	strErr += "Error message: ";
	strErr += m_strError;
	return strErr;
}

wxString TrackAction::GetDisplayStr() const
{
	wxString str = c_strAction[m_action];
	str += ":\n";
	str += m_pTrack[0]->GetDisplayName();
	return str;
}



void TrackAction::SetDefault() 
{
	if (m_bDone || HadError() || m_bUserModified) return;
	if (m_pTrack[0]->IsFaulty())
	{
		m_action = Delete;
		SetChoice(0, Delete);
		SetChoice(1, noAction);
		m_nChoices = 2;
		return;
	}
	if (m_pTrack[0]->MissingTag())
	{	// not allowed transfer tracks with no id3 tags
		m_action = noAction;
		SetChoice(0, Delete);
		SetChoice(1, noAction);
		m_nChoices = 2;
		return;
	}
	if (m_pTrack[1] == NULL)	// single action
	{
		SetChoice(0, Delete);
		SetChoice(2, noAction);
		m_nChoices = 3;
		if (m_pTrack[0]->IsNomad())
		{
			SetChoice(1, Nomad2PC);
		}
		else
		{
			SetChoice(1, PC2Nomad);
		}
		if (m_pDependAction)
		{
			if ((m_pDependAction->m_action == PC2Nomad || 
				m_pDependAction->m_action == Nomad2PC) &&
				m_pTrack[0]->EqualNoCase(m_pDependAction->m_pTrack[0]))
			{
				m_action = Delete;
			}
			else m_action = noAction;
			return;
		}
		if (m_pTrack[0]->IsNomad())
		{
			m_action = m_config.GetActionDefault(c_nDefMissingOnPC);
		}
		else
		{
			m_action = m_config.GetActionDefault(c_nDefMissingOnNomad);
		}
	}
	else	// TrackAction has 2 tracks
	{
		m_action = m_config.GetActionDefault(c_nDefPair);
		SetChoice(0, Label2Nomad);
		SetChoice(1, Label2PC);
		SetChoice(2, PC2Nomad);
		SetChoice(3, Delete);
		SetChoice(4, noAction);
		m_nChoices = 5;
	}
} // Set Default


// trackIX = 1 when it is the 2nd row in a label action
wxString TrackAction::GetValue(int nCol, int nTrackIX) const
{
	if (nCol == c_nActionCol) 
	{
		// to do: if nTrackIX == 1 then it?s impossble for m_pTrack[1] == NULL
		if (m_bExecuting) return "Running";
		if (nTrackIX == 1 && m_pTrack[1] != NULL) return "";
		if (HadError()) 
		{
			wxString strErr = "Error: " + m_strError;
			return strErr;
		}
		if (!m_bDone) return c_strAction[m_action];
		return c_strActionDone[m_action];
	}
	return m_pTrack[nTrackIX]->GetValue(nCol);
}


wxGridCellAttr* TrackAction::GetAttr(int nCol, int nTrackIX, CellAttrCache& attrCache) const
{
	if (nCol == c_nIconCol) 
	{
		return attrCache.Get((CellTypeUsed)(ctIcon+GetColourIndex() ) );
	}
	if (nCol == c_nActionCol)
	{
		if (m_bExecuting) return attrCache.Get((CellTypeUsed)ctNormal);
		if ((nTrackIX == 1 && m_pTrack[1] != NULL) || m_bDone || HadError()) 
			return attrCache.Get((CellTypeUsed)(ctNormal+GetColourIndex()));
		if (m_pAttr) 
		{
			m_pAttr->IncRef();
			return m_pAttr;
		}
		m_pAttr = new wxGridCellAttr;
		m_pAttr->IncRef();
		if (m_bUserModified) m_pAttr->SetTextColour(*wxBLUE);
		m_pAttr->SetBackgroundColour(wxSystemSettings::GetSystemColour(wxSYS_COLOUR_WINDOW));
		wxLayoutAlignment border = wxLAYOUT_NONE;
		m_pAttr->SetRenderer(new wxGridCellChoiceRenderer(border));
		m_pAttr->SetEditor(new wxFastComboEditor(m_nChoices, m_strChoices, TRUE));
		return m_pAttr;
	}
	return attrCache.Get((CellTypeUsed)(ctNormal+GetColourIndex()));
}	// GetAttr


// for icon column we should sort according to pc, nomad & pair
bool TrackAction::ColIsLessThan(int nCol, const Action* pAction) const
{
	TrackAction* pTrackAction = (TrackAction*)pAction;
	if (nCol == c_nActionCol) 
	{
		if (c_strAction[m_action] == c_strAction[pTrackAction->GetAction()]) 
			return m_pTrack[0]->ColIsLessThan(c_nTitleCol, pTrackAction->GetTrack(0));
		return c_strAction[m_action] < c_strAction[pTrackAction->GetAction()]; 
	}
	if (nCol == c_nIconCol) 
	{
		if (GetTypeOrder() == pTrackAction->GetTypeOrder())
			return m_pTrack[0]->ColIsLessThan(c_nTitleCol, pTrackAction->GetTrack(0));
		return GetTypeOrder() < pTrackAction->GetTypeOrder();
	}
	return m_pTrack[0]->ColIsLessThan(nCol, pTrackAction->GetTrack(0));
}

int TrackAction::GetTypeOrder() const	
{
	if (m_pTrack[1] != NULL) return 4;
	if (m_pDependAction != NULL) return 3;
	if (m_pTrack[0]->IsNomad()) return 2;
	return 1;
}

bool TrackAction::IsPaired() const	
{
	if (m_pTrack[1] != NULL || m_pDependAction != NULL)
		return true;
	return false;
}


