#include "Precompiled.h"
#include "TrackMatcher.h"
#include "Config.h"
#include "Track.h"
#ifdef __WINDOWS__
	#include "WNomad.h"
#else
	#include "LNomad.h"
#endif
#include "TrackAction.h"
#include "EzMessageBox.h"
#include "ActionTable.h"
#include "TrackCache.h"

void Matcher::RemoveAll()
{
	m_mapPCTrack.clear();
	m_mapExtPCTrack.clear();
	m_mapNomadTrack.clear();
	m_listDup.clear();
	m_pTrackCache->RemoveHits();
}

// TraverseDirs: makes use of the base class wxDirTraverser
// to traverse the directory hierarchy looking for mp3
void Matcher::TraverseDirs(const wxString& strExt)
{
	wxDir dirSearch;
	int n;
	// traverse each search directory 
	for (n = 0; n < c_nIncludeDirs; n++)
	{
		if (m_config.IsValidDir(n))
		{
			dirSearch.Open(m_config.GetDir(n));
			dirSearch.Traverse(*this, strExt);
#ifndef __WINDOWS__
			// linux is case sensitive, therefore we have to expressly search for
			// .MP3 as well as .mp3
			dirSearch.Traverse(*this, strExt.Upper());	
#endif
		}
	}
}

// OnFile is called when we find a mp3 when doing a directory traverse
wxDirTraverseResult Matcher::OnFile(const wxString& strFileSpec)
{
	// Work around for Bug in FindFirstFile: returns files ending in tilda!
	if (strFileSpec.Last() == '~') return wxDIR_CONTINUE;
	Track* pTrack = m_pTrackCache->GetFile(strFileSpec);
	wxString strKey;		
	strKey = pTrack->GetKey();
	TrackMap::iterator iterMap; 
	iterMap = m_mapPCTrack.find(strKey); 
	if (iterMap != m_mapPCTrack.end()) 
	{
		wxLogMessage("Duplicate track ignored:");
		wxLogMessage(pTrack->GetDisplayName());
		m_listDup.push_back(pTrack);
	}
	else
	{
		m_mapPCTrack[strKey] = pTrack;
		strKey = pTrack->GetKey(&m_config);
		m_mapExtPCTrack[strKey] = pTrack;
	}
	return wxDIR_CONTINUE;
}	//	OnFile


// OnDir: used to ignore directories on the exclude list
wxDirTraverseResult Matcher::OnDir(const wxString& strDir)
{
	if (m_config.IsExcluded(strDir)) return wxDIR_IGNORE;
	return wxDIR_CONTINUE;
}

// LoadNomadData: loads tracks and playlists from the nomad
// if pConfig is null then the track key is generated without genre, etc. embedded 
void Matcher::LoadNomadData(Nomad* pNomad, const Config* pConfig)
{
	Track* pTrack;
	TrackMap::iterator iterMap; 
	wxString strKey;		
	bool bMore = pNomad->StartLoadAllTracks();
	while (bMore)
	{
		pTrack = pNomad->LoadTrackData();
		if (pTrack)
		{
			if (pTrack->GetType() == typeMP3)
			{
				strKey = pTrack->GetKey(pConfig);
				iterMap = m_mapNomadTrack.find(strKey); 
				if (iterMap != m_mapNomadTrack.end()) 
				{
					wxLogMessage("Duplicate track ignored:");
					wxLogMessage(pTrack->GetDisplayName());
				}
				else m_mapNomadTrack[strKey] = pTrack;
			}
			else
			{
				wxLogMessage("Non-MP3 ignored: %s!", 
					(const char*)pTrack->GetDisplayName());
			}
		}
		bMore = pNomad->MoveNextTrack();
	}
	pNomad->LoadPlaylists();
	if (pNomad->HadError())
	{
		wxLogMessage("Error reading tracks: %s!", (const char*)pNomad->GetLastErrorStr());
		EzMessageBox("Error reading jukebox data", pNomad->GetLastErrorStr(), 
			0, "OK");
	}
}	//	LoadNomadData


//-------------------------------------------------------------------------------------

void TrackMatcher::RemoveAll()
{
	Matcher::RemoveAll();
	m_listDiffSize.clear();
	m_listPCExtra.clear();
	m_listRelabeled.clear();
}

void TrackMatcher::Run(Nomad* pNomad)
{
	RemoveAll();
	TraverseDirs("*.mp3");
	LoadNomadData(pNomad, &m_config);
	Match();
}

// Match tracks with the same key and files size 
// results in m_mapNomadTrack containing all nomad tracks not found on PC
// and m_listPCExtra containing all PC tracks not found on Nomad
// m_listDiffSize contains tracks with same key and different files sizes 
// m_listRelabeled contains tracks with same files sizes and slightly different labels
void TrackMatcher::Match()
{
	wxLogMessage("Tracks on Nomad %d, Tracks on PC %d, dups %d", 
		m_mapNomadTrack.size(), m_mapPCTrack.size(),
		m_listDup.size());
	wxString strKey;
	TrackMap::iterator iterMapNomad; 
	TrackMap::const_iterator iterMapPC; 
	for (iterMapPC = m_mapExtPCTrack.begin(); iterMapPC != m_mapExtPCTrack.end(); iterMapPC++) 
	{
		strKey = iterMapPC->second->GetKey(&m_config);
		iterMapNomad = m_mapNomadTrack.find(strKey); 
		if (iterMapNomad != m_mapNomadTrack.end())
		{
			if (!iterMapPC->second->SameSizeAndLen(iterMapNomad->second) )
				m_listDiffSize.push_back(
					TrackPair(iterMapPC->second, iterMapNomad->second));
			m_mapNomadTrack.erase(iterMapNomad);
		}
		else
			m_listPCExtra.push_back(iterMapPC->second);
	}
	// Match tracks for relabel
	// we match in decreaseing match strictness, to reduce the amount of bad matches
	for (int nSimLevel = 1; nSimLevel <= c_nMaxSimLevel; nSimLevel++)
	{
		FindLabelMatches(m_listRelabeled, nSimLevel);
	}
	FindLabelMatches(m_listDiffSize, c_nSameLabelsNoCase);
	wxLogMessage("Non-matched tracks, %d on Nomad, %d  on PC", 
		m_listPCExtra.size(), m_mapNomadTrack.size());
	wxLogMessage("Similiar tracks %d", m_listRelabeled.size()+m_listDiffSize.size());
}	//	Match


void TrackMatcher::FindLabelMatches(vector<TrackPair>& listPair, int nSimLevel)
{
	TrackMap::iterator iterMap, iterTemp; 
	vector<Track*>::iterator iterVector;
	iterMap = m_mapNomadTrack.begin();
	bool bMatch;
	while (iterMap != m_mapNomadTrack.end()) 
	{
		bMatch = false;
		for (iterVector = m_listPCExtra.begin(); iterVector != m_listPCExtra.end(); 
			iterVector++) 
		{
			if (iterMap->second->IsSimiliar(*iterVector, nSimLevel))
			{
				listPair.push_back(TrackPair(*iterVector, iterMap->second));
				// we must store the iterator, because it will become invalid when
				// we call erase
				iterTemp = iterMap;
				iterTemp++;
				m_mapNomadTrack.erase(iterMap);
				iterMap = iterTemp;
				m_listPCExtra.erase(iterVector);
				bMatch = true;
				break;
			}
		}	// end for PCExtra
		if (!bMatch) iterMap++;		// it will already have gone forward if match
	}
}	 //	FindLabelMatches


void TrackMatcher::CreateActions(ActionTable& table) const
{
	unsigned int n;
	TrackAction *pAction, *pAction2;

	table.RemoveActions();
	for (n = 0; n < m_listPCExtra.size(); n++)
	{
		pAction = new TrackAction(m_config);
		pAction->SetTrack(0, m_listPCExtra[n]);
		pAction->SetDefault();
		table.AddAction(pAction);
	}

	TrackMap::const_iterator iterMap; 
	for (iterMap = m_mapNomadTrack.begin(); iterMap != m_mapNomadTrack.end(); iterMap++) 
	{
		pAction = new TrackAction(m_config);
		pAction->SetTrack(0, iterMap->second);
		pAction->SetDefault();
		table.AddAction(pAction);
	}

	for (n = 0; n < m_listDiffSize.size(); n++)
	{
		pAction = new TrackAction(m_config);
		pAction->SetTrack(0, m_listDiffSize[n].m_pTrackPC);
		pAction2 = new TrackAction(m_config);
		pAction2->SetTrack(0, m_listDiffSize[n].m_pTrackNomad);
		pAction2->SetDependAction(pAction);
		pAction->SetDependAction(pAction2);
		pAction->SetDefault();
		pAction2->SetDefault();
		table.AddAction(pAction);
		table.AddAction(pAction2);
	}
	for (n = 0; n < m_listRelabeled.size(); n++)
	{
		pAction = new TrackAction(m_config);
		pAction->SetTrack(0, m_listRelabeled[n].m_pTrackPC);
		pAction->SetTrack(1, m_listRelabeled[n].m_pTrackNomad);
		pAction->SetDefault();
		table.AddAction(pAction);
	}
	table.Sort(c_nIconCol);
}	// end CreateActions


